/*
 * RealTimeSchedule.hh
 *
 * Copyright (c) 2003, 2004 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/**
 * @file RealTimeSchedule.hh
 *
 * Header file for the RealTimeSchedule helper class.
 */

#ifndef _cpu_schedule_hh
#define _cpu_schedule_hh

#include "BrokerC.h"

#include <rt_scheduler.h>

/**
 * The RealTimeSchedule class is used to help policies that manage
 * Broker::RealTimeTask's.
 */
class RealTimeSchedule
{

public:

    /**
     * Construct a RealTimeSchedule object with the given values.
     *
     * @param start The start time for the task.
     * @param period The task period in microseconds.
     * @param deadline The task deadline in microseconds, relative to the start
     * of the period.
     * @param reserve_mode The reservation mode for the schedule.
     * @param pid A PID for a process to be scheduled.
     */
    RealTimeSchedule(CORBA::ULong start = 0,
		     CORBA::ULong period = 0,
		     CORBA::ULong deadline = 0,
		     rts_reserve_mode_t reserve_mode = RTS_RSV_SOFT,
		     pid_t pid = -1);

    /**
     * Set the members from a Broker::ScheduleParameters sequence.
     *
     * @param sp The sequence to take values from.
     */
    void setParameters(const Broker::ScheduleParameters &sp)
	throw (Broker::DuplicateScheduleParameter,
	       Broker::InvalidScheduleParameter,
	       Broker::MissingScheduleParameter);
    
    /**
     * Deconstruct a RealTimeSchedule object.
     */
    virtual ~RealTimeSchedule();

    /**
     * Compute the number of microseconds that this schedule will have to
     * execute during the given time frame.
     *
     * @param start The start of the time frame, in microseconds.
     * @param stop The end of the time frame, in microseconds.
     * @return The number of microseconds during the time frame in which this
     *         task will be schedulable.
     */
    CORBA::ULong intersect(CORBA::ULong start, CORBA::ULong stop);

    /**
     * The start time for the task.
     *
     * @todo Use a long long or timespec here.
     */
    CORBA::ULong rts_Start;
    
    /** The task period in microseconds. */
    CORBA::ULong rts_Period;
    
    /**
     * The task deadline in microseconds, relative to the start of the
     * period.
     */
    CORBA::ULong rts_Deadline;

    /**
     * The reservation mode for the schedule, either hard or soft.
     */
    rts_reserve_mode_t rts_ReserveMode;

    /**
     * The process ID this schedule refers to.
     */
    pid_t rts_PID;

};

#endif
